@@ -4,6 +4,8 @@
 
 #include "system.h"
 
+#include <string.h>
+
 #include <rpm/rpmcli.h>
 #include <rpm/rpmtag.h>
 #include <rpm/rpmlib.h>		/* rpmReadPackageFile, vercmp etc */
@@ -14,8 +16,8 @@
 #include <rpm/rpmlog.h>
 #include <rpm/rpmfileutil.h>
 
-#include "lib/rpmgi.h"
-#include "lib/manifest.h"
+#include "rpmgi.h"
+#include "manifest.h"
 #include "debug.h"
 
 static int rpmcliPackagesTotal = 0;
@@ -83,12 +85,6 @@ static rpmVSFlags setvsFlags(struct rpmInstallArguments_s * ia)
     else
 	vsflags = rpmExpandNumeric("%{?_vsflags_install}");
 
-    if (rpmcliQueryFlags & VERIFY_DIGEST)
-	vsflags |= _RPMVSF_NODIGESTS;
-    if (rpmcliQueryFlags & VERIFY_SIGNATURE)
-	vsflags |= _RPMVSF_NOSIGNATURES;
-    if (rpmcliQueryFlags & VERIFY_HDRCHK)
-	vsflags |= RPMVSF_NOHDRCHK;
     vsflags |= rpmcliVSFlags;
 
     return vsflags;
@@ -256,10 +252,10 @@ struct rpmEIU {
     rpmRC rpmrc;
 };
 
-static int rpmcliTransaction(rpmts ts, struct rpmInstallArguments_s * ia,
-		      int numPackages)
+static int rpmcliTransaction(rpmts ts, struct rpmInstallArguments_s * ia)
 {
     rpmps ps;
+    int numPackages = rpmtsNElements(ts);
 
     int rc = 0;
     int stop = 0;
@@ -291,8 +287,15 @@ static int rpmcliTransaction(rpmts ts, struct rpmInstallArguments_s * ia,
     }
 
     if (numPackages && !stop) {
-	rpmlog(RPMLOG_DEBUG, eflags ? "erasing packages\n" :
-				      "installing binary packages\n");
+	const char *msg;
+	if (ia->installInterfaceFlags & INSTALL_RESTORE)
+	    msg = "restoring packages";
+	else if (ia->installInterfaceFlags & INSTALL_ERASE)
+	    msg = "erasing packages";
+	else
+	    msg = "installing binary packages";
+
+	rpmlog(RPMLOG_DEBUG, "%s\n", msg);
 	rpmtsClean(ts);
 	rc = rpmtsRun(ts, NULL, ia->probFilter);
 
@@ -352,7 +355,6 @@ static int tryReadHeader(rpmts ts, struct rpmEIU * eiu, Header * hdrp)
 	   fd = NULL;
        }
        eiu->numFailed++;
-       *eiu->fnp = NULL;
        return RPMRC_FAIL;
    }
 
@@ -368,7 +370,6 @@ static int tryReadHeader(rpmts ts, struct rpmEIU * eiu, Header * hdrp)
    if (eiu->rpmrc == RPMRC_FAIL) {
        rpmlog(RPMLOG_ERR, _("%s cannot be installed\n"), *eiu->fnp);
        eiu->numFailed++;
-       *eiu->fnp = NULL;
    }
 
    return RPMRC_OK;
@@ -376,19 +377,21 @@ static int tryReadHeader(rpmts ts, struct rpmEIU * eiu, Header * hdrp)
 
 
 /* On --freshen, verify package is installed and newer */
-static int checkFreshenStatus(rpmts ts, Header h)
+static int checkFreshenStatus(rpmts ts, Header h, int allowOld)
 {
     rpmdbMatchIterator mi = NULL;
     const char * name = headerGetString(h, RPMTAG_NAME);
     const char *arch = headerGetString(h, RPMTAG_ARCH);
     Header oldH = NULL;
-
     if (name != NULL)
         mi = rpmtsInitIterator(ts, RPMDBI_NAME, name, 0);
     if (rpmtsColor(ts) && arch)
 	rpmdbSetIteratorRE(mi, RPMTAG_ARCH, RPMMIRE_DEFAULT, arch);
 
     while ((oldH = rpmdbNextIterator(mi)) != NULL) {
+	/* Package is older than those currently installed. */
+	if (allowOld && rpmVersionCompare(oldH, h) > 0)
+	    break;
 	/* Package is newer than those currently installed. */
         if (rpmVersionCompare(oldH, h) < 0)
 	    break;
@@ -414,7 +417,7 @@ static int rpmNoGlob(const char *fn, int *argcPtr, ARGV_t * argvPtr)
 /** @todo Generalize --freshen policies. */
 int rpmInstall(rpmts ts, struct rpmInstallArguments_s * ia, ARGV_t fileArgv)
 {
-    struct rpmEIU * eiu = xcalloc(1, sizeof(*eiu));
+    struct rpmEIU * eiu = (struct rpmEIU *)xcalloc(1, sizeof(*eiu));
     rpmRelocation * relocations;
     char * fileURL = NULL;
     rpmVSFlags vsflags, ovsflags;
@@ -456,9 +459,7 @@ int rpmInstall(rpmts ts, struct rpmInstallArguments_s * ia, ARGV_t fileArgv)
 	if (giFlags & RPMGI_NOGLOB) {
 	    rc = rpmNoGlob(*eiu->fnp, &ac, &av);
 	} else {
-	    char * fn = rpmEscapeSpaces(*eiu->fnp);
-	    rc = rpmGlob(fn, &ac, &av);
-	    fn = _free(fn);
+	    rc = rpmGlobPath(*eiu->fnp, RPMGLOB_NOCHECK, &ac, &av);
 	}
 	if (rc || ac == 0) {
 	    if (giFlags & RPMGI_NOGLOB) {
@@ -467,6 +468,7 @@ int rpmInstall(rpmts ts, struct rpmInstallArguments_s * ia, ARGV_t fileArgv)
 		rpmlog(RPMLOG_ERR, _("File not found by glob: %s\n"), *eiu->fnp);
 	    }
 	    eiu->numFailed++;
+	    argvFree(av);
 	    continue;
 	}
 
@@ -518,7 +520,10 @@ restart:
 			_("skipping %s - transfer failed\n"), fileURL);
 		eiu->numFailed++;
 		eiu->pkgURL[eiu->pkgx] = NULL;
-		tfn = _free(tfn);
+		if (tfn) {
+		    (void) unlink(tfn);
+		    tfn = _free(tfn);
+		}
 		break;
 	    }
 	    eiu->pkgState[eiu->pkgx] = 1;
@@ -550,20 +555,29 @@ restart:
 	rpmlog(RPMLOG_DEBUG, "============== %s\n", *eiu->fnp);
 	(void) urlPath(*eiu->fnp, &fileName);
 
-	if (tryReadHeader(ts, eiu, &h) == RPMRC_FAIL)
+	if (tryReadHeader(ts, eiu, &h) == RPMRC_FAIL) {
+	    if (eiu->pkgState[eiu->fnp - eiu->pkgURL] == 1)
+		(void) unlink(*eiu->fnp);
+	    *eiu->fnp = _free(*eiu->fnp);
 	    continue;
+	}
 
 	if (eiu->rpmrc == RPMRC_NOTFOUND) {
 	    rc = tryReadManifest(eiu);
 	    if (rc == RPMRC_OK) {
 	        eiu->prevx++;
+		headerFree(h);
 	        goto restart;
 	    }
+	} else if (eiu->rpmrc == RPMRC_FAIL) {
+	    if (eiu->pkgState[eiu->fnp - eiu->pkgURL] == 1)
+		(void) unlink(*eiu->fnp);
+	    *eiu->fnp = _free(*eiu->fnp);
 	}
 
 	if (headerIsSource(h)) {
+	    headerFree(h);
 	    if (ia->installInterfaceFlags & INSTALL_FRESHEN) {
-		headerFree(h);
 	        continue;
 	    }
 	    rpmlog(RPMLOG_DEBUG, "\tadded source package [%d]\n",
@@ -588,12 +602,13 @@ restart:
 		rpmlog(RPMLOG_ERR, _("package %s is not relocatable\n"),
 		       headerGetString(h, RPMTAG_NAME));
 		eiu->numFailed++;
+		headerFree(h);
 		goto exit;
 	    }
 	}
 
 	if (ia->installInterfaceFlags & INSTALL_FRESHEN)
-	    if (checkFreshenStatus(ts, h) != 1) {
+	    if (checkFreshenStatus(ts, h, ia->probFilter & RPMPROB_FILTER_OLDPACKAGE) != 1) {
 		headerFree(h);
 	        continue;
 	    }
@@ -635,7 +650,7 @@ restart:
     if (eiu->numFailed) goto exit;
 
     if (eiu->numRPMS) {
-        int rc = rpmcliTransaction(ts, ia, eiu->numPkgs);
+        int rc = rpmcliTransaction(ts, ia);
         if (rc < 0)
             eiu->numFailed += eiu->numRPMS;
 	else if (rc > 0)
@@ -647,7 +662,6 @@ restart:
 	rpmcliProgressTotal = 0;
 	rpmcliProgressCurrent = 0;
 	for (i = 0; i < eiu->numSRPMS; i++) {
-	    rpmsqPoll();
 	    if (eiu->sourceURL[i] != NULL) {
 	        rc = RPMRC_OK;
 		if (!(rpmtsFlags(ts) & RPMTRANS_FLAG_TEST))
@@ -667,8 +681,17 @@ exit:
 	    eiu->pkgURL[i] = _free(eiu->pkgURL[i]);
 	}
     }
+    if (eiu->sourceURL != NULL) {
+        for (i = 0; i < eiu->numSRPMS; i++) {
+	    if (eiu->sourceURL[i] == NULL) continue;
+	    if (eiu->pkgState[i] == 1)
+	        (void) unlink(eiu->sourceURL[i]);
+	    eiu->sourceURL[i] = _free(eiu->sourceURL[i]);
+	}
+    }
     eiu->pkgState = _free(eiu->pkgState);
     eiu->pkgURL = _free(eiu->pkgURL);
+    eiu->sourceURL = _free(eiu->sourceURL);
     eiu->argv = _free(eiu->argv);
     rc = eiu->numFailed;
     free(eiu);
@@ -733,7 +756,7 @@ int rpmErase(rpmts ts, struct rpmInstallArguments_s * ia, ARGV_const_t argv)
     free(qfmt);
 
     if (numFailed) goto exit;
-    numFailed = rpmcliTransaction(ts, ia, numPackages);
+    numFailed = rpmcliTransaction(ts, ia);
 exit:
     rpmtsEmpty(ts);
     rpmtsSetVSFlags(ts, ovsflags);
@@ -741,6 +764,35 @@ exit:
     return (numFailed < 0) ? numPackages : numFailed;
 }
 
+static int handleRestorePackage(QVA_t qva, rpmts ts, Header h)
+{
+    return rpmtsAddRestoreElement(ts, h);
+}
+
+int rpmRestore(rpmts ts, struct rpmInstallArguments_s * ia, ARGV_const_t argv)
+{
+    QVA_t qva = &rpmQVKArgs;
+    int rc;
+    rpmVSFlags vsflags = setvsFlags(ia);
+    rpmVSFlags ovsflags = rpmtsSetVSFlags(ts, vsflags);
+
+    rpmtsSetFlags(ts, ia->transFlags);
+    setNotifyFlag(ia, ts);
+
+    if (qva->qva_showPackage == NULL)
+	qva->qva_showPackage = handleRestorePackage;
+
+    rc = rpmcliArgIter(ts, qva, argv);
+    if (rc == 0) {
+	rc = rpmcliTransaction(ts, ia);
+    }
+
+    rpmtsEmpty(ts);
+    rpmtsSetVSFlags(ts, ovsflags);
+
+    return rc;
+}
+
 int rpmInstallSource(rpmts ts, const char * arg,
 		char ** specFilePtr, char ** cookie)
 {
