@@ -3,6 +3,8 @@
 #include <sys/types.h>
 #include <sys/wait.h>
 #include <errno.h>
+#include <fcntl.h>
+#include <signal.h>
 
 #include <rpm/rpmfileutil.h>
 #include <rpm/rpmmacro.h>
@@ -11,16 +13,17 @@
 #include <rpm/header.h>
 #include <rpm/rpmds.h>
 
-#include "rpmio/rpmlua.h"
-#include "lib/rpmscript.h"
-#include "rpmio/rpmio_internal.h"
+#include "rpmlua.h"
+#include "rpmscript.h"
+#include "rpmio_internal.h"
 
-#include "lib/rpmplugins.h"     /* rpm plugins hooks */
+#include "rpmchroot.h"
+#include "rpmplugins.h"     /* rpm plugins hooks */
 
 #include "debug.h"
 
 struct scriptNextFileFunc_s {
-    char *(*func)(void *);	/* function producing input for script */
+    nextfilefunc func;		/* function producing input for script */
     void *param;		/* parameter for func */
 };
 
@@ -33,7 +36,8 @@ struct rpmScript_s {
     char *body;			/* script body */
     char *descr;		/* description for logging */
     rpmscriptFlags flags;	/* flags to control operation */
-    struct scriptNextFileFunc_s nextFileFunc;  /* input function */
+    int chroot;			/* chrooted script? */
+    struct scriptNextFileFunc_s *nextFileFunc;  /* input function */
 };
 
 struct scriptInfo_s {
@@ -47,39 +51,48 @@ struct scriptInfo_s {
 };
 
 static const struct scriptInfo_s scriptInfo[] = {
-    { RPMSCRIPT_PREIN, "%prein", 0,
+    { RPMSCRIPT_PREIN, "prein", 0,
 	RPMTAG_PREIN, RPMTAG_PREINPROG, RPMTAG_PREINFLAGS,
 	RPMSCRIPT_FLAG_CRITICAL, },
-    { RPMSCRIPT_PREUN, "%preun", 0,
+    { RPMSCRIPT_PREUN, "preun", 0,
 	RPMTAG_PREUN, RPMTAG_PREUNPROG, RPMTAG_PREUNFLAGS,
 	RPMSCRIPT_FLAG_CRITICAL, },
-    { RPMSCRIPT_POSTIN, "%post", 0,
+    { RPMSCRIPT_POSTIN, "post", 0,
 	RPMTAG_POSTIN, RPMTAG_POSTINPROG, RPMTAG_POSTINFLAGS,
 	0, },
-    { RPMSCRIPT_POSTUN, "%postun", 0,
+    { RPMSCRIPT_POSTUN, "postun", 0,
 	RPMTAG_POSTUN, RPMTAG_POSTUNPROG, RPMTAG_POSTUNFLAGS,
 	0, },
-    { RPMSCRIPT_PRETRANS, "%pretrans", 0,
+    { RPMSCRIPT_PRETRANS, "pretrans", 0,
 	RPMTAG_PRETRANS, RPMTAG_PRETRANSPROG, RPMTAG_PRETRANSFLAGS,
 	RPMSCRIPT_FLAG_CRITICAL, },
-    { RPMSCRIPT_POSTTRANS, "%posttrans", 0,
+    { RPMSCRIPT_POSTTRANS, "posttrans", 0,
 	RPMTAG_POSTTRANS, RPMTAG_POSTTRANSPROG, RPMTAG_POSTTRANSFLAGS,
 	0, },
-    { RPMSCRIPT_TRIGGERPREIN, "%triggerprein", RPMSENSE_TRIGGERPREIN,
+    { RPMSCRIPT_PREUNTRANS, "preuntrans", 0,
+	RPMTAG_PREUNTRANS, RPMTAG_PREUNTRANSPROG, RPMTAG_PREUNTRANSFLAGS,
+	RPMSCRIPT_FLAG_CRITICAL, },
+    { RPMSCRIPT_POSTUNTRANS, "postuntrans", 0,
+	RPMTAG_POSTUNTRANS, RPMTAG_POSTUNTRANSPROG, RPMTAG_POSTUNTRANSFLAGS,
+	0, },
+    { RPMSCRIPT_TRIGGERPREIN, "triggerprein", RPMSENSE_TRIGGERPREIN,
 	RPMTAG_TRIGGERPREIN, 0, 0,
 	0, },
-    { RPMSCRIPT_TRIGGERUN, "%triggerun", RPMSENSE_TRIGGERUN,
+    { RPMSCRIPT_TRIGGERUN, "triggerun", RPMSENSE_TRIGGERUN,
 	RPMTAG_TRIGGERUN, 0, 0,
 	0, },
-    { RPMSCRIPT_TRIGGERIN, "%triggerin", RPMSENSE_TRIGGERIN,
+    { RPMSCRIPT_TRIGGERIN, "triggerin", RPMSENSE_TRIGGERIN,
 	RPMTAG_TRIGGERIN, 0, 0,
 	0, },
-    { RPMSCRIPT_TRIGGERPOSTUN, "%triggerpostun", RPMSENSE_TRIGGERPOSTUN,
+    { RPMSCRIPT_TRIGGERPOSTUN, "triggerpostun", RPMSENSE_TRIGGERPOSTUN,
 	RPMTAG_TRIGGERPOSTUN, 0, 0,
 	0, },
-    { RPMSCRIPT_VERIFY, "%verify", 0,
+    { RPMSCRIPT_VERIFY, "verify", 0,
 	RPMTAG_VERIFYSCRIPT, RPMTAG_VERIFYSCRIPTPROG, RPMTAG_VERIFYSCRIPTFLAGS,
 	RPMSCRIPT_FLAG_CRITICAL, },
+    { RPMSCRIPT_SYSUSERS, "sysusers", 0,
+	RPMTAG_SYSUSERS, 0, 0,
+	RPMSCRIPT_FLAG_CRITICAL, },
     { 0, "unknown", 0,
 	RPMTAG_NOT_FOUND, RPMTAG_NOT_FOUND, RPMTAG_NOT_FOUND,
 	0, }
@@ -92,6 +105,29 @@ static const struct scriptInfo_s * findTag(rpmTagVal tag)
 	si++;
     return si;
 }
+
+static int next_file(lua_State *L)
+{
+    scriptNextFileFunc nff = (scriptNextFileFunc)lua_touserdata(L, lua_upvalueindex(1));
+    lua_pushstring(L, nff->func(nff->param));
+    return 1;
+}
+
+int rpmScriptChrootIn(rpmScript script)
+{
+    int rc = 0;
+    if (script->chroot)
+	rc = rpmChrootIn();
+    return rc;
+}
+
+int rpmScriptChrootOut(rpmScript script)
+{
+    int rc = 0;
+    if (script->chroot)
+	rc = rpmChrootOut();
+    return rc;
+}
 /**
  * Run internal Lua script.
  */
@@ -100,52 +136,55 @@ static rpmRC runLuaScript(rpmPlugins plugins, ARGV_const_t prefixes,
 		   ARGV_t * argvp, const char *script, int arg1, int arg2,
 		   scriptNextFileFunc nextFileFunc)
 {
+    char *scriptbuf = NULL;
     rpmRC rc = RPMRC_FAIL;
-#ifdef WITH_LUA
-    ARGV_t argv = argvp ? *argvp : NULL;
-    rpmlua lua = NULL; /* Global state. */
-    rpmluav var = rpmluavNew();
+    rpmlua lua = rpmluaGetGlobalState();
+    lua_State *L = (lua_State *)rpmluaGetLua(lua);
     int cwd = -1;
 
     rpmlog(RPMLOG_DEBUG, "%s: running <lua> scriptlet.\n", sname);
 
-    /* Create arg variable */
-    rpmluaPushTable(lua, "arg");
-    rpmluavSetListMode(var, 1);
-    rpmluaSetNextFileFunc(nextFileFunc->func, nextFileFunc->param);
-    if (argv) {
-	char **p;
-	for (p = argv; *p; p++) {
-	    rpmluavSetValue(var, RPMLUAV_STRING, *p);
-	    rpmluaSetVar(lua, var);
+    if (nextFileFunc) {
+	lua_getglobal(L, "rpm");
+	lua_pushlightuserdata(L, nextFileFunc);
+	lua_pushcclosure(L, &next_file, 1);
+	lua_setfield(L, -2, "next_file");
+    }
+
+    if (prefixes) {
+	lua_newtable(L);
+	for (ARGV_const_t p = prefixes; p && *p; p++) {
+	    int i = p - prefixes + 1;
+	    lua_pushstring(L, *p);
+	    lua_rawseti(L, -2, i);
 	}
+	lua_setglobal(L, "RPM_INSTALL_PREFIX");
     }
-    if (arg1 >= 0) {
-	rpmluavSetValueNum(var, arg1);
-	rpmluaSetVar(lua, var);
+
+    if (arg1 >= 0)
+	argvAddNum(argvp, arg1);
+    if (arg2 >= 0)
+	argvAddNum(argvp, arg2);
+
+    if (arg1 >= 0 || arg2 >= 0) {
+	/* Hack to convert arguments to numbers for backwards compat. Ugh. */
+	rstrscat(&scriptbuf,
+		arg1 >= 0 ? "arg[2] = tonumber(arg[2]);" : "",
+		arg2 >= 0 ? "arg[3] = tonumber(arg[3]);" : "",
+		script, NULL);
+        script = scriptbuf;
     }
-    if (arg2 >= 0) {
-	rpmluavSetValueNum(var, arg2);
-	rpmluaSetVar(lua, var);
-    }
-    rpmluaPop(lua);
 
     /* Lua scripts can change our cwd and umask, save and restore */
-    /* XXX TODO: use cwd from chroot state to save unnecessary open here */
     cwd = open(".", O_RDONLY);
     if (cwd != -1) {
 	mode_t oldmask = umask(0);
 	umask(oldmask);
-	pid_t pid = getpid();
 
-	if (chdir("/") == 0 && rpmluaRunScript(lua, script, sname) == 0) {
+	if (chdir("/") == 0 &&
+		rpmluaRunScript(lua, script, sname, NULL, *argvp) == 0) {
 	    rc = RPMRC_OK;
 	}
-	if (pid != getpid()) {
-	    /* Terminate child process forked in lua scriptlet */
-	    rpmlog(RPMLOG_ERR, _("No exec() called after fork() in lua scriptlet\n"));
-	    _exit(EXIT_FAILURE);
-	}
 	/* This failing would be fatal, return something different for it... */
 	if (fchdir(cwd)) {
 	    rpmlog(RPMLOG_ERR, _("Unable to restore current directory: %m"));
@@ -154,13 +193,18 @@ static rpmRC runLuaScript(rpmPlugins plugins, ARGV_const_t prefixes,
 	close(cwd);
 	umask(oldmask);
     }
+    free(scriptbuf);
 
-    rpmluaDelVar(lua, "arg");
-    rpmluavFree(var);
+    if (prefixes) {
+	lua_pushnil(L);
+	lua_setglobal(L, "RPM_INSTALL_PREFIX");
+    }
 
-#else
-    rpmlog(lvl, _("<lua> scriptlet support not built in\n"));
-#endif
+    if (nextFileFunc) {
+	lua_pushnil(L);
+	lua_setfield(L, -2, "next_file");
+	lua_pop(L, 1); /* "rpm" global */
+    }
 
     return rc;
 }
@@ -171,9 +215,19 @@ static void doScriptExec(ARGV_const_t argv, ARGV_const_t prefixes,
 			FD_t scriptFd, FD_t out)
 {
     int xx;
+    struct sigaction act;
+    sigset_t set;
+
+    /* Unmask most signals, the scripts may need them */
+    sigfillset(&set);
+    sigdelset(&set, SIGINT);
+    sigdelset(&set, SIGQUIT);
+    sigprocmask(SIG_UNBLOCK, &set, NULL);
 
     /* SIGPIPE is ignored in rpm, reset to default for the scriptlet */
-    (void) signal(SIGPIPE, SIG_DFL);
+    memset(&act, 0, sizeof(act));
+    act.sa_handler = SIG_DFL;
+    sigaction(SIGPIPE, &act, NULL);
 
     rpmSetCloseOnExec();
 
@@ -216,9 +270,6 @@ static void doScriptExec(ARGV_const_t argv, ARGV_const_t prefixes,
     }
 	
     if (chdir("/") == 0) {
-	/* XXX Don't mtrace into children. */
-	unsetenv("MALLOC_CHECK_");
-
 	xx = execv(argv[0], argv);
 	if (xx) {
 	    rpmlog(RPMLOG_ERR,
@@ -265,7 +316,7 @@ static rpmRC runExtScript(rpmPlugins plugins, ARGV_const_t prefixes,
     char * fn = NULL;
     pid_t pid, reaped;
     int status;
-    int inpipe[2];
+    int inpipe[2] = { -1, -1 };
     FILE *in = NULL;
     const char *line;
     char *mline = NULL;
@@ -339,7 +390,7 @@ static rpmRC runExtScript(rpmPlugins plugins, ARGV_const_t prefixes,
     close(inpipe[0]);
     inpipe[0] = 0;
 
-    if (nextFileFunc->func) {
+    if (nextFileFunc) {
 	while ((line = nextFileFunc->func(nextFileFunc->param)) != NULL) {
 	    size_t size = strlen(line);
 	    size_t ret_size;
@@ -431,9 +482,9 @@ rpmRC rpmScriptRun(rpmScript script, int arg1, int arg2, FD_t scriptFd,
 
     if (rc != RPMRC_FAIL) {
 	if (script_type & RPMSCRIPTLET_EXEC) {
-	    rc = runExtScript(plugins, prefixes, script->descr, lvl, scriptFd, &args, script->body, arg1, arg2, &script->nextFileFunc);
+	    rc = runExtScript(plugins, prefixes, script->descr, lvl, scriptFd, &args, script->body, arg1, arg2, script->nextFileFunc);
 	} else {
-	    rc = runLuaScript(plugins, prefixes, script->descr, lvl, scriptFd, &args, script->body, arg1, arg2, &script->nextFileFunc);
+	    rc = runLuaScript(plugins, prefixes, script->descr, lvl, scriptFd, &args, script->body, arg1, arg2, script->nextFileFunc);
 	}
     }
 
@@ -471,41 +522,40 @@ static const char * tag2sln(rpmTagVal tag)
 }
 
 static rpmScript rpmScriptNew(Header h, rpmTagVal tag, const char *body,
-			      rpmscriptFlags flags)
+			      rpmscriptFlags flags, const char *prefix)
 {
     char *nevra = headerGetAsString(h, RPMTAG_NEVRA);
-    rpmScript script = xcalloc(1, sizeof(*script));
+    rpmScript script = (rpmScript)xcalloc(1, sizeof(*script));
     script->tag = tag;
     script->type = getScriptType(tag);
     script->flags = getDefFlags(tag) | flags;
     script->body = (body != NULL) ? xstrdup(body) : NULL;
-    rasprintf(&script->descr, "%s(%s)", tag2sln(tag), nevra);
+    script->chroot = 1;
+    rasprintf(&script->descr, "%%%s%s(%s)", prefix, tag2sln(tag), nevra);
 
     /* macros need to be expanded before possible queryformat */
     if (script->body && (script->flags & RPMSCRIPT_FLAG_EXPAND)) {
-	char *body = rpmExpand(script->body, NULL);
+	char *b = rpmExpand(script->body, NULL);
 	free(script->body);
-	script->body = body;
+	script->body = b;
     }
     if (script->body && (script->flags & RPMSCRIPT_FLAG_QFORMAT)) {
 	/* XXX TODO: handle queryformat errors */
-	char *body = headerFormat(h, script->body, NULL);
+	char *b = headerFormat(h, script->body, NULL);
 	free(script->body);
-	script->body = body;
+	script->body = b;
     }
 
-    script->nextFileFunc.func = NULL;
-    script->nextFileFunc.param = NULL;
-
     free(nevra);
     return script;
 }
 
-void rpmScriptSetNextFileFunc(rpmScript script, char *(*func)(void *),
+void rpmScriptSetNextFileFunc(rpmScript script, nextfilefunc func,
 			    void *param)
 {
-    script->nextFileFunc.func = func;
-    script->nextFileFunc.param = param;
+    script->nextFileFunc = (struct scriptNextFileFunc_s *)xmalloc(sizeof(*script->nextFileFunc));
+    script->nextFileFunc->func = func;
+    script->nextFileFunc->param = param;
 }
 
 rpmTagVal triggerDsTag(rpmscriptTriggerModes tm)
@@ -570,6 +620,7 @@ rpmScript rpmScriptFromTriggerTag(Header h, rpmTagVal triggerTag,
     rpmScript script = NULL;
     struct rpmtd_s tscripts, tprogs, tflags;
     headerGetFlags hgflags = HEADERGET_MINMEM;
+    const char *prefix = "";
 
     switch (tm) {
 	case RPMSCRIPT_NORMALTRIGGER:
@@ -581,12 +632,16 @@ rpmScript rpmScriptFromTriggerTag(Header h, rpmTagVal triggerTag,
 	    headerGet(h, RPMTAG_FILETRIGGERSCRIPTS, &tscripts, hgflags);
 	    headerGet(h, RPMTAG_FILETRIGGERSCRIPTPROG, &tprogs, hgflags);
 	    headerGet(h, RPMTAG_FILETRIGGERSCRIPTFLAGS, &tflags, hgflags);
+	    prefix = "file";
 	    break;
 	case RPMSCRIPT_TRANSFILETRIGGER:
 	    headerGet(h, RPMTAG_TRANSFILETRIGGERSCRIPTS, &tscripts, hgflags);
 	    headerGet(h, RPMTAG_TRANSFILETRIGGERSCRIPTPROG, &tprogs, hgflags);
 	    headerGet(h, RPMTAG_TRANSFILETRIGGERSCRIPTFLAGS, &tflags, hgflags);
+	    prefix = "transfile";
 	    break;
+	default:
+	    return NULL;
     }
 
     if (rpmtdSetIndex(&tscripts, ix) >= 0 && rpmtdSetIndex(&tprogs, ix) >= 0) {
@@ -596,10 +651,11 @@ rpmScript rpmScriptFromTriggerTag(Header h, rpmTagVal triggerTag,
 	if (rpmtdSetIndex(&tflags, ix) >= 0)
 	    sflags = rpmtdGetNumber(&tflags);
 
-	script = rpmScriptNew(h, triggerTag, rpmtdGetString(&tscripts), sflags);
+	script = rpmScriptNew(h, triggerTag,
+				rpmtdGetString(&tscripts), sflags, prefix);
 
 	/* hack up a hge-style NULL-terminated array */
-	script->args = xmalloc(2 * sizeof(*script->args) + strlen(prog) + 1);
+	script->args = (char **)xmalloc(2 * sizeof(*script->args) + strlen(prog) + 1);
 	script->args[0] = (char *)(script->args + 2);
 	script->args[1] = NULL;
 	strcpy(script->args[0], prog);
@@ -612,6 +668,19 @@ rpmScript rpmScriptFromTriggerTag(Header h, rpmTagVal triggerTag,
     return script;
 }
 
+rpmScript rpmScriptFromArgv(Header h, rpmTagVal scriptTag, ARGV_t argv, rpmscriptFlags flags, int chroot)
+{
+    rpmScript script = NULL;
+
+    if (h && argv) {
+	char *body = argvJoin(argv, " ");
+	script = rpmScriptNew(h, scriptTag, body, flags, "");
+	script->chroot = chroot;
+	free(body);
+    }
+    return script;
+}
+
 rpmScript rpmScriptFromTag(Header h, rpmTagVal scriptTag)
 {
     rpmScript script = NULL;
@@ -622,10 +691,10 @@ rpmScript rpmScriptFromTag(Header h, rpmTagVal scriptTag)
 
 	script = rpmScriptNew(h, scriptTag,
 			      headerGetString(h, scriptTag),
-			      headerGetNumber(h, getFlagTag(scriptTag)));
+			      headerGetNumber(h, getFlagTag(scriptTag)), "");
 
 	if (headerGet(h, progTag, &prog, (HEADERGET_ALLOC|HEADERGET_ARGV))) {
-	    script->args = prog.data;
+	    script->args = (char **)prog.data;
 	}
     }
     return script;
@@ -637,6 +706,7 @@ rpmScript rpmScriptFree(rpmScript script)
 	free(script->args);
 	free(script->body);
 	free(script->descr);
+	free(script->nextFileFunc);
 	free(script);
     }
     return NULL;
