@@ -14,7 +14,7 @@
 #include <rpm/rpmlog.h>
 #include <rpm/argv.h>
 
-#include "rpmio/rpmio_internal.h"
+#include "rpmio_internal.h"
 
 #include "debug.h"
 
@@ -25,7 +25,7 @@ int rpmDoDigest(int algo, const char * fn,int asAscii, unsigned char * digest)
 {
     unsigned char * dig = NULL;
     size_t diglen, buflen = 32 * BUFSIZ;
-    unsigned char *buf = xmalloc(buflen);
+    unsigned char *buf = (unsigned char *)xmalloc(buflen);
     int rc = 0;
 
     FD_t fd = Fopen(fn, "r.ufdio");
@@ -188,7 +188,7 @@ int rpmFileIsCompressed(const char * file, rpmCompressedMagic * compressed)
 	       (magic[4] == 0x5a) && (magic[5] == 0x00)) {
 	/* new style xz (lzma) with magic */
 	*compressed = COMPRESSED_XZ;
-    } else if ((magic[0] == 0x28) && (magic[1] == 0x85) &&
+    } else if ((magic[0] == 0x28) && (magic[1] == 0xB5) &&
 	       (magic[2] == 0x2f)                     ) {
 	*compressed = COMPRESSED_ZSTD;
     } else if ((magic[0] == 'L') && (magic[1] == 'Z') &&
@@ -227,10 +227,8 @@ char *rpmCleanPath(char * path)
     if (path == NULL)
 	return NULL;
 
-/*fprintf(stderr, "*** RCP %s ->\n", path); */
     s = t = te = path;
     while (*s != '\0') {
-/*fprintf(stderr, "*** got \"%.*s\"\trest \"%s\"\n", (t-path), path, s); */
 	switch (*s) {
 	case ':':			/* handle url's */
 	    if (s[1] == '/' && s[2] == '/') {
@@ -246,7 +244,6 @@ char *rpmCleanPath(char * path)
 		{};
 	    if (se < t && *se == '/') {
 		te = se;
-/*fprintf(stderr, "*** next pdir \"%.*s\"\n", (te-path), path); */
 	    }
 	    while (s[1] == '/')
 		s++;
@@ -261,7 +258,6 @@ char *rpmCleanPath(char * path)
 	    /* as "../.", and the last '.' is stripped.  This   */
 	    /* would not be correct processing.                 */
 	    if (begin && s[1] == '.' && (s[2] == '/' || s[2] == '\0')) {
-/*fprintf(stderr, "    leading \"..\"\n"); */
 		*t++ = *s++;
 		break;
 	    }
@@ -289,7 +285,6 @@ char *rpmCleanPath(char * path)
 		if (te > path)
 		    for (--te; te > path && *te != '/'; te--)
 			{};
-/*fprintf(stderr, "*** prev pdir \"%.*s\"\n", (te-path), path); */
 		s++;
 		s++;
 		continue;
@@ -307,7 +302,6 @@ char *rpmCleanPath(char * path)
 	t--;
     *t = '\0';
 
-/*fprintf(stderr, "\t%s\n", path); */
     return path;
 }
 
@@ -386,7 +380,7 @@ char * rpmGetPath(const char *path, ...)
     return rpmCleanPath(res);
 }
 
-char * rpmEscapeSpaces(const char * s)
+static char * rpmEscapeChars(const char *s, const char *accept, int (*fn)(int))
 {
     const char * se;
     char * t;
@@ -394,15 +388,15 @@ char * rpmEscapeSpaces(const char * s)
     size_t nb = 0;
 
     for (se = s; *se; se++) {
-	if (isspace(*se))
+	if ((accept && strchr(accept, *se)) || (fn && fn(*se)))
 	    nb++;
 	nb++;
     }
     nb++;
 
-    t = te = xmalloc(nb);
+    t = te = (char *)xmalloc(nb);
     for (se = s; *se; se++) {
-	if (isspace(*se))
+	if ((accept && strchr(accept, *se)) || (fn && fn(*se)))
 	    *te++ = '\\';
 	*te++ = *se;
     }
@@ -410,6 +404,34 @@ char * rpmEscapeSpaces(const char * s)
     return t;
 }
 
+char * rpmEscapeSpaces(const char *s)
+{
+    return rpmEscapeChars(s, NULL, isspace);
+}
+
+char * rpmEscape(const char *s, const char *accept)
+{
+    return rpmEscapeChars(s, accept, NULL);
+}
+
+void rpmUnescape(char *s, const char *accept)
+{
+    char *p, *q;
+    int skip = 0;
+
+    p = q = s;
+    while (*q != '\0') {
+	*p = *q++;
+	if (*p == '\\' && (!accept || strchr(accept, *q)) && !skip) {
+	    skip = 1;
+	    continue;
+	}
+	p++;
+	skip = 0;
+    }
+    *p = '\0';
+}
+
 int rpmFileHasSuffix(const char *path, const char *suffix)
 {
     size_t plen = strlen(path);
@@ -439,6 +461,8 @@ int rpmMkdirs(const char *root, const char *pathstr)
     
     for (char **d = dirs; *d; d++) {
 	char *path = rpmGetPath(root ? root : "", *d, NULL);
+	if (strstr(path, "%{"))
+	    rpmlog(RPMLOG_WARNING, ("undefined macro(s) in %s: %s\n"), *d, path);
 	if ((rc = rpmioMkpath(path, 0755, -1, -1)) != 0) {
 	    const char *msg = _("failed to create directory");
 	    /* try to be more informative if the failing part was a macro */
@@ -458,7 +482,7 @@ int rpmMkdirs(const char *root, const char *pathstr)
 static void setConfigDir(void)
 {
     char *rpmenv = getenv("RPM_CONFIGDIR");
-    rpm_config_dir = rpmenv ? xstrdup(rpmenv) : RPMCONFIGDIR;
+    rpm_config_dir = rpmenv ? xstrdup(rpmenv) : RPM_CONFIGDIR;
 }
 
 const char *rpmConfigDir(void)
