@@ -8,7 +8,14 @@
 #include <errno.h>
 #include <sys/wait.h>
 #include <popt.h>
-#include <libgen.h>
+#include <fcntl.h>
+#ifdef WITH_FSVERITY
+#include <libfsverity.h>
+#include "rpmsignverity.h"
+#endif
+#ifdef WITH_IMAEVM
+#include "rpmsignfiles.h"
+#endif
 
 #include <rpm/rpmlib.h>			/* RPMSIGTAG & related */
 #include <rpm/rpmmacro.h>
@@ -17,12 +24,11 @@
 #include <rpm/rpmfileutil.h>	/* rpmMkTemp() */
 #include <rpm/rpmlog.h>
 #include <rpm/rpmstring.h>
-#include <rpmio/rpmio_internal.h>
+#include "rpmio_internal.h"
 
-#include "lib/rpmlead.h"
-#include "lib/signature.h"
-#include "lib/rpmvs.h"
-#include "sign/rpmsignfiles.h"
+#include "rpmlead.h"
+#include "signature.h"
+#include "rpmvs.h"
 
 #include "debug.h"
 
@@ -33,68 +39,6 @@ typedef struct sigTarget_s {
     rpm_loff_t size;
 } *sigTarget;
 
-/*
- * There is no function for creating unique temporary fifos so create
- * unique temporary directory and then create fifo in it.
- */
-static char *mkTempFifo(void)
-{
-    char *tmppath = NULL, *tmpdir = NULL, *fifofn = NULL;
-    mode_t mode;
-
-    tmppath = rpmExpand("%{_tmppath}", NULL);
-    if (rpmioMkpath(tmppath, 0755, (uid_t) -1, (gid_t) -1))
-	goto exit;
-
-
-    tmpdir = rpmGetPath(tmppath, "/rpm-tmp.XXXXXX", NULL);
-    mode = umask(0077);
-    tmpdir = mkdtemp(tmpdir);
-    umask(mode);
-    if (tmpdir == NULL) {
-	rpmlog(RPMLOG_ERR, _("error creating temp directory %s: %m\n"),
-	    tmpdir);
-	tmpdir = _free(tmpdir);
-	goto exit;
-    }
-
-    fifofn = rpmGetPath(tmpdir, "/fifo", NULL);
-    if (mkfifo(fifofn, 0600) == -1) {
-	rpmlog(RPMLOG_ERR, _("error creating fifo %s: %m\n"), fifofn);
-	fifofn = _free(fifofn);
-    }
-
-exit:
-    if (fifofn == NULL && tmpdir != NULL)
-	unlink(tmpdir);
-
-    free(tmppath);
-    free(tmpdir);
-
-    return fifofn;
-}
-
-/* Delete fifo and then temporary directory in which it was located */
-static int rpmRmTempFifo(const char *fn)
-{
-    int rc = 0;
-    char *dfn = NULL, *dir = NULL;
-
-    if ((rc = unlink(fn)) != 0) {
-	rpmlog(RPMLOG_ERR, _("error delete fifo %s: %m\n"), fn);
-	return rc;
-    }
-
-    dfn = xstrdup(fn);
-    dir = dirname(dfn);
-
-    if ((rc = rmdir(dir)) != 0)
-	rpmlog(RPMLOG_ERR, _("error delete directory %s: %m\n"), dir);
-    free(dfn);
-
-    return rc;
-}
-
 static int closeFile(FD_t *fdp)
 {
     if (fdp == NULL || *fdp == NULL)
@@ -187,7 +131,7 @@ exit:
 
 /*
  * Validate generated signature and insert to header if it looks sane.
- * NSS doesn't support everything GPG does. Basic tests to see if the 
+ * RPM doesn't support everything GPG does. Basic tests to see if the
  * generated signature is something we can use.
  * Return generated signature tag data on success, NULL on failure.
  */
@@ -198,33 +142,42 @@ static rpmtd makeSigTag(Header sigh, int ishdr, uint8_t *pkt, size_t pktlen)
     rpmtd sigtd = NULL;
     unsigned int hash_algo;
     unsigned int pubkey_algo;
+    int ver;
 
     if (pgpPrtParams(pkt, pktlen, PGPTAG_SIGNATURE, &sigp)) {
-	rpmlog(RPMLOG_ERR, _("Unsupported PGP signature\n"));
+	rpmlog(RPMLOG_ERR, _("Unsupported OpenPGP signature\n"));
 	goto exit;
     }
 
     hash_algo = pgpDigParamsAlgo(sigp, PGPVAL_HASHALGO);
     if (rpmDigestLength(hash_algo) == 0) {
-	rpmlog(RPMLOG_ERR, _("Unsupported PGP hash algorithm %u\n"), hash_algo);
+	rpmlog(RPMLOG_ERR, _("Unsupported OpenPGP hash algorithm %u\n"), hash_algo);
 	goto exit;
     }
 
     pubkey_algo = pgpDigParamsAlgo(sigp, PGPVAL_PUBKEYALGO);
     switch (pubkey_algo) {
     case PGPPUBKEYALGO_DSA:
+    case PGPPUBKEYALGO_ECDSA:
+    case PGPPUBKEYALGO_EDDSA:
 	sigtag = ishdr ? RPMSIGTAG_DSA : RPMSIGTAG_GPG;
 	break;
     case PGPPUBKEYALGO_RSA:
 	sigtag = ishdr ? RPMSIGTAG_RSA : RPMSIGTAG_PGP;
 	break;
     default:
-	rpmlog(RPMLOG_ERR, _("Unsupported PGP pubkey algorithm %u\n"),
+	rpmlog(RPMLOG_ERR, _("Unsupported OpenPGP pubkey algorithm %u\n"),
 		pubkey_algo);
 	goto exit;
 	break;
     }
 
+    ver = pgpDigParamsVersion(sigp);
+    if (ver < 4) {
+	rpmlog(RPMLOG_WARNING, _("Deprecated OpenPGP signature version %d\n"),
+		ver);
+    }
+
     /* Looks sane, create the tag data */
     sigtd = rpmtdNew();
     sigtd->count = pktlen;
@@ -238,47 +191,80 @@ exit:
     return sigtd;
 }
 
+static char ** signCmd(const char *sigfile)
+{
+    int argc = 0;
+    char **argv = NULL;
+
+    rpmPushMacro(NULL, "__plaintext_filename", NULL, "-", -1);
+    rpmPushMacro(NULL, "__signature_filename", NULL, sigfile, -1);
+
+    char *cmd = rpmExpand("%{?__gpg_sign_cmd}", NULL);
+
+    rpmPopMacro(NULL, "__plaintext_filename");
+    rpmPopMacro(NULL, "__signature_filename");
+
+    if (poptParseArgvString(cmd, &argc, (const char ***)&argv) < 0 || argc < 2) {
+	rpmlog(RPMLOG_ERR, _("Invalid sign command: %s\n"), cmd);
+	argv = _free(argv);
+    }
+
+    free(cmd);
+
+    return argv;
+}
+
 static int runGPG(sigTarget sigt, const char *sigfile)
 {
     int pid = 0, status;
-    FD_t fnamedPipe = NULL;
-    char *namedPipeName = NULL;
+    int pipefd[2];
+    FILE *fpipe = NULL;
     unsigned char buf[BUFSIZ];
     ssize_t count;
     ssize_t wantCount;
     rpm_loff_t size;
     int rc = 1; /* assume failure */
+    char **argv = NULL;
 
-    namedPipeName = mkTempFifo();
+    if ((argv = signCmd(sigfile)) == NULL)
+	goto exit_nowait;
 
-    rpmPushMacro(NULL, "__plaintext_filename", NULL, namedPipeName, -1);
-    rpmPushMacro(NULL, "__signature_filename", NULL, sigfile, -1);
+    if (pipe(pipefd) < 0) {
+        rpmlog(RPMLOG_ERR, _("Could not create pipe for signing: %m\n"));
+        goto exit_nowait;
+    }
 
     if (!(pid = fork())) {
-	char *const *av;
-	char *cmd = NULL;
-	const char *gpg_path = rpmExpand("%{?_gpg_path}", NULL);
+	/* GnuPG needs extra setup, try to see if that's what we're running */
+	char *out = rpmExpand("%(", argv[0], " --version 2> /dev/null)", NULL);
+	int using_gpg = (strstr(out, "GnuPG") != NULL);
+	if (using_gpg) {
+	    const char *tty = ttyname(STDIN_FILENO);
+	    const char *gpg_path = NULL;
 
-	if (gpg_path && *gpg_path != '\0')
-	    (void) setenv("GNUPGHOME", gpg_path, 1);
+	    if (!getenv("GPG_TTY") && (!tty || setenv("GPG_TTY", tty, 0)))
+		rpmlog(RPMLOG_WARNING, _("Could not set GPG_TTY to stdin: %m\n"));
 
-	unsetenv("MALLOC_CHECK_");
-	cmd = rpmExpand("%{?__gpg_sign_cmd}", NULL);
-	rc = poptParseArgvString(cmd, NULL, (const char ***)&av);
-	if (!rc)
-	    rc = execve(av[0], av+1, environ);
+	    gpg_path = rpmExpand("%{?_gpg_path}", NULL);
+	    if (gpg_path && *gpg_path != '\0')
+		(void) setenv("GNUPGHOME", gpg_path, 1);
+	}
+	free(out);
 
-	rpmlog(RPMLOG_ERR, _("Could not exec %s: %s\n"), "gpg",
-			strerror(errno));
+	dup2(pipefd[0], STDIN_FILENO);
+	close(pipefd[1]);
+
+	rc = execve(argv[0], argv+1, environ);
+
+	rpmlog(RPMLOG_ERR, _("Could not exec %s: %s\n"), argv[0],
+		strerror(errno));
 	_exit(EXIT_FAILURE);
     }
 
-    rpmPopMacro(NULL, "__plaintext_filename");
-    rpmPopMacro(NULL, "__signature_filename");
-
-    fnamedPipe = Fopen(namedPipeName, "w");
-    if (!fnamedPipe) {
-	rpmlog(RPMLOG_ERR, _("Fopen failed\n"));
+    close(pipefd[0]);
+    fpipe = fdopen(pipefd[1], "w");
+    if (!fpipe) {
+	rpmlog(RPMLOG_ERR, _("Could not open pipe for writing: %m\n"));
 	goto exit;
     }
 
@@ -291,8 +277,8 @@ static int runGPG(sigTarget sigt, const char *sigfile)
     size = sigt->size;
     wantCount = size < sizeof(buf) ? size : sizeof(buf);
     while ((count = Fread(buf, sizeof(buf[0]), wantCount, sigt->fd)) > 0) {
-	Fwrite(buf, sizeof(buf[0]), count, fnamedPipe);
-	if (Ferror(fnamedPipe)) {
+	fwrite(buf, sizeof(buf[0]), count, fpipe);
+	if (ferror(fpipe)) {
 	    rpmlog(RPMLOG_ERR, _("Could not write to pipe\n"));
 	    goto exit;
 	}
@@ -304,29 +290,31 @@ static int runGPG(sigTarget sigt, const char *sigfile)
 		sigt->fileName, Fstrerror(sigt->fd));
 	goto exit;
     }
-    Fclose(fnamedPipe);
-    fnamedPipe = NULL;
 
-    (void) waitpid(pid, &status, 0);
-    pid = 0;
-    if (!WIFEXITED(status) || WEXITSTATUS(status)) {
-	rpmlog(RPMLOG_ERR, _("gpg exec failed (%d)\n"), WEXITSTATUS(status));
+exit:
+
+    if (fpipe)
+	fclose(fpipe);
+    if (pipefd[1])
+	close(pipefd[1]);
+
+    pid_t reaped;
+    do {
+        reaped = waitpid(pid, &status, 0);
+    } while (reaped == -1 && errno == EINTR);
+
+    if (reaped == -1) {
+	rpmlog(RPMLOG_ERR, _("%s waitpid failed (%s)\n"), argv[0],
+		strerror(errno));
+    } else if (!WIFEXITED(status) || WEXITSTATUS(status)) {
+	rpmlog(RPMLOG_ERR, _("%s exec failed (%d)\n"), argv[0],
+		WEXITSTATUS(status));
     } else {
 	rc = 0;
     }
 
-exit:
-
-    if (fnamedPipe)
-	Fclose(fnamedPipe);
-
-    if (pid)
-	waitpid(pid, &status, 0);
-
-    if (namedPipeName) {
-	rpmRmTempFifo(namedPipeName);
-	free(namedPipeName);
-    }
+exit_nowait:
+    free(argv);
 
     return rc;
 }
@@ -351,14 +339,14 @@ static rpmtd makeGPGSignature(Header sigh, int ishdr, sigTarget sigt)
 	goto exit;
 
     if (stat(sigfile, &st)) {
-	/* GPG failed to write signature */
-	rpmlog(RPMLOG_ERR, _("gpg failed to write signature\n"));
+	/* External command failed to write signature */
+	rpmlog(RPMLOG_ERR, _("failed to write signature\n"));
 	goto exit;
     }
 
     pktlen = st.st_size;
-    rpmlog(RPMLOG_DEBUG, "GPG sig size: %zd\n", pktlen);
-    pkt = xmalloc(pktlen);
+    rpmlog(RPMLOG_DEBUG, "OpenPGP sig size: %zd\n", pktlen);
+    pkt = (uint8_t *)xmalloc(pktlen);
 
     {	FD_t fd;
 
@@ -374,7 +362,7 @@ static rpmtd makeGPGSignature(Header sigh, int ishdr, sigTarget sigt)
 	}
     }
 
-    rpmlog(RPMLOG_DEBUG, "Got %zd bytes of GPG sig\n", pktlen);
+    rpmlog(RPMLOG_DEBUG, "Got %zd bytes of OpenPGP sig\n", pktlen);
 
     /* Parse the signature, change signature tag as appropriate. */
     sigtd = makeSigTag(sigh, ishdr, pkt, pktlen);
@@ -395,6 +383,14 @@ static void deleteSigs(Header sigh)
     headerDel(sigh, RPMSIGTAG_PGP5);
 }
 
+static void deleteFileSigs(Header sigh)
+{
+    headerDel(sigh, RPMSIGTAG_FILESIGNATURELENGTH);
+    headerDel(sigh, RPMSIGTAG_FILESIGNATURES);
+    headerDel(sigh, RPMSIGTAG_VERITYSIGNATURES);
+    headerDel(sigh, RPMSIGTAG_VERITYSIGNATUREALGO);
+}
+
 static int haveSignature(rpmtd sigtd, Header h)
 {
     pgpDigParams sig1 = NULL;
@@ -405,12 +401,12 @@ static int haveSignature(rpmtd sigtd, Header h)
     if (!headerGet(h, rpmtdTag(sigtd), &oldtd, HEADERGET_DEFAULT))
 	return rc;
 
-    pgpPrtParams(sigtd->data, sigtd->count, PGPTAG_SIGNATURE, &sig1);
+    pgpPrtParams((uint8_t *)sigtd->data, sigtd->count, PGPTAG_SIGNATURE, &sig1);
     while (rpmtdNext(&oldtd) >= 0 && rc == 0) {
-	pgpPrtParams(oldtd.data, oldtd.count, PGPTAG_SIGNATURE, &sig2);
+	pgpPrtParams((uint8_t *)oldtd.data, oldtd.count, PGPTAG_SIGNATURE, &sig2);
 	if (pgpDigParamsCmp(sig1, sig2) == 0)
 	    rc = 1;
-	pgpDigParamsFree(sig2);
+	sig2 = pgpDigParamsFree(sig2);
     }
     pgpDigParamsFree(sig1);
     rpmtdFreeData(&oldtd);
@@ -437,14 +433,17 @@ static int replaceSignature(Header sigh, sigTarget sigt_v3, sigTarget sigt_v4)
 
     if (headerPut(sigh, sigtd, HEADERPUT_DEFAULT) == 0)
 	goto exit;
-    rpmtdFree(sigtd);
 
-    /* Assume the same signature test holds for v3 signature too */
-    if ((sigtd = makeGPGSignature(sigh, 0, sigt_v3)) == NULL)
-	goto exit;
+    if (sigt_v3) {
+	rpmtdFree(sigtd);
 
-    if (headerPut(sigh, sigtd, HEADERPUT_DEFAULT) == 0)
-	goto exit;
+	/* Assume the same signature test holds for v3 signature too */
+	if ((sigtd = makeGPGSignature(sigh, 0, sigt_v3)) == NULL)
+	    goto exit;
+
+	if (headerPut(sigh, sigtd, HEADERPUT_DEFAULT) == 0)
+	    goto exit;
+    }
 
     rc = 0;
 exit:
@@ -455,11 +454,19 @@ exit:
 static void unloadImmutableRegion(Header *hdrp, rpmTagVal tag)
 {
     struct rpmtd_s td;
+    Header oh = NULL;
 
     if (headerGet(*hdrp, tag, &td, HEADERGET_DEFAULT)) {
-	Header oh = headerCopyLoad(td.data);
-	Header nh = headerCopy(oh);
+	oh = headerImport(td.data, td.count, HEADERIMPORT_COPY);
 	rpmtdFreeData(&td);
+    } else {
+	/* XXX should we warn if the immutable region is corrupt/missing? */
+	oh = headerLink(*hdrp);
+    }
+
+    if (oh) {
+	/* Perform a copy to eliminate crud from buggy signing tools etc */
+	Header nh = headerCopy(oh);
 	headerFree(*hdrp);
 	*hdrp = headerLink(nh);
 	headerFree(nh);
@@ -490,25 +497,72 @@ static rpmRC includeFileSignatures(Header *sigp, Header *hdrp)
 #endif
 }
 
+static rpmRC includeVeritySignatures(FD_t fd, Header *sigp, Header *hdrp)
+{
+#ifdef WITH_FSVERITY
+    rpmRC rc = RPMRC_OK;
+    char *key = rpmExpand("%{?_file_signing_key}", NULL);
+    char *keypass = rpmExpand("%{?_file_signing_key_password}", NULL);
+    char *cert = rpmExpand("%{?_file_signing_cert}", NULL);
+    char *algorithm = rpmExpand("%{?_verity_algorithm}", NULL);
+    uint16_t algo = 0;
+
+    if (rstreq(keypass, "")) {
+	free(keypass);
+	keypass = NULL;
+    }
+
+    if (algorithm && strlen(algorithm) > 0) {
+	    algo = libfsverity_find_hash_alg_by_name(algorithm);
+	    rpmlog(RPMLOG_DEBUG, _("Searching for algorithm %s got %i\n"),
+		   algorithm, algo);
+	    if (!algo) {
+		    rpmlog(RPMLOG_ERR, _("Unsupported fsverity algorithm %s\n"),
+			   algorithm);
+		    rc = RPMRC_FAIL;
+		    goto out;
+	    }
+    }
+    if (key && cert) {
+	    rc = rpmSignVerity(fd, *sigp, *hdrp, key, keypass, cert, algo);
+    } else {
+	rpmlog(RPMLOG_ERR, _("fsverity signatures requires a key and a cert\n"));
+	rc = RPMRC_FAIL;
+    }
+
+ out:
+    free(keypass);
+    free(key);
+    free(cert);
+    return rc;
+#else
+    rpmlog(RPMLOG_ERR, _("fsverity signing support not built in\n"));
+    return RPMRC_FAIL;
+#endif
+}
+
 static int msgCb(struct rpmsinfo_s *sinfo, void *cbdata)
 {
-    char **msg = cbdata;
+    char **msg = (char **)cbdata;
     if (sinfo->rc && *msg == NULL)
 	*msg = rpmsinfoMsg(sinfo);
     return (sinfo->rc != RPMRC_FAIL);
 }
 
 /* Require valid digests on entire package for signing. */
-static int checkPkg(FD_t fd, char **msg)
+static rpmRC checkPkg(FD_t fd, char **msg)
 {
-    int rc;
+    rpmRC rc;
     struct rpmvs_s *vs = rpmvsCreate(RPMSIG_DIGEST_TYPE, 0, NULL);
     off_t offset = Ftell(fd);
 
     Fseek(fd, 0, SEEK_SET);
     rc = rpmpkgRead(vs, fd, NULL, NULL, msg);
-    if (!rc)
-	rc = rpmvsVerify(vs, RPMSIG_DIGEST_TYPE, msgCb, msg);
+    if (!rc) {
+	rc = rpmvsVerify(vs, RPMSIG_DIGEST_TYPE, msgCb, msg) ?
+			RPMRC_FAIL : RPMRC_OK;
+    }
+    Fseek(fd, offset, SEEK_SET);
     Fseek(fd, offset, SEEK_SET);
 
     rpmvsFree(vs);
@@ -519,10 +573,10 @@ static int checkPkg(FD_t fd, char **msg)
  * Create/modify elements in signature header.
  * @param rpm		path to package
  * @param deleting	adding or deleting signature?
- * @param signfiles	sign files if non-zero
+ * @param flags
  * @return		0 on success, -1 on error
  */
-static int rpmSign(const char *rpm, int deleting, int signfiles)
+static int rpmSign(const char *rpm, int deleting, int flags)
 {
     FD_t fd = NULL;
     FD_t ofd = NULL;
@@ -575,18 +629,32 @@ static int rpmSign(const char *rpm, int deleting, int signfiles)
 	goto exit;
     }
 
+    /* Always add V3 signatures if no payload digest present */
+    if (!(headerIsEntry(h, RPMTAG_PAYLOADDIGEST) ||
+	  headerIsEntry(h, RPMTAG_PAYLOADDIGESTALT))) {
+	flags |= RPMSIGN_FLAG_RPMV3;
+    }
+
     unloadImmutableRegion(&sigh, RPMTAG_HEADERSIGNATURES);
     origSigSize = headerSizeof(sigh, HEADER_MAGIC_YES);
 
-    if (signfiles) {
+    if (flags & RPMSIGN_FLAG_IMA) {
 	if (includeFileSignatures(&sigh, &h))
 	    goto exit;
     }
 
-    if (deleting) {	/* Nuke all the signature tags. */
+    if (flags & RPMSIGN_FLAG_FSVERITY) {
+	if (includeVeritySignatures(fd, &sigh, &h))
+	    goto exit;
+    }
+
+    if (deleting == 2) {	/* Nuke IMA + fsverity file signature tags. */
+	deleteFileSigs(sigh);
+    } else if (deleting) {	/* Nuke all the signature tags. */
 	deleteSigs(sigh);
     } else {
 	/* Signature target containing header + payload */
+	int v3 = (flags & RPMSIGN_FLAG_RPMV3);
 	sigt_v3.fd = fd;
 	sigt_v3.start = headerStart;
 	sigt_v3.fileName = rpm;
@@ -596,7 +664,7 @@ static int rpmSign(const char *rpm, int deleting, int signfiles)
 	sigt_v4 = sigt_v3;
 	sigt_v4.size = headerSizeof(h, HEADER_MAGIC_YES);
 
-	res = replaceSignature(sigh, &sigt_v3, &sigt_v4);
+	res = replaceSignature(sigh, v3 ? &sigt_v3 : NULL, &sigt_v4);
 	if (res != 0) {
 	    if (res == 1) {
 		rpmlog(RPMLOG_WARNING,
@@ -610,14 +678,20 @@ static int rpmSign(const char *rpm, int deleting, int signfiles)
 	res = -1;
     }
 
-    /* Try to make new signature smaller to have size of original signature */
+    /* Adjust reserved size for added/removed signatures */
     if (headerGet(sigh, RPMSIGTAG_RESERVEDSPACE, &utd, HEADERGET_MINMEM)) {
 	int diff = headerSizeof(sigh, HEADER_MAGIC_YES) - origSigSize;
 
-	if (diff > 0 && diff < utd.count) {
+	/* diff can be zero if nothing was added or removed */
+	if (diff) {
 	    utd.count -= diff;
-	    headerMod(sigh, &utd);
-	    insSig = 1;
+	    if (utd.count > 0 && utd.count < origSigSize) {
+		uint8_t *zeros = (uint8_t *)xcalloc(utd.count, sizeof(*zeros));
+		utd.data = zeros;
+		headerMod(sigh, &utd);
+		insSig = 1;
+		free(zeros);
+	    }
 	}
     }
 
@@ -626,14 +700,14 @@ static int rpmSign(const char *rpm, int deleting, int signfiles)
     if (sigh == NULL)	/* XXX can't happen */
 	goto exit;
 
+    if (Fseek(fd, insSig ? sigStart : 0, SEEK_SET) < 0) {
+	rpmlog(RPMLOG_ERR, _("Could not seek in file %s: %s\n"),
+		rpm, Fstrerror(fd));
+	goto exit;
+    }
+
     if (insSig) {
 	/* Insert new signature into original rpm */
-	if (Fseek(fd, sigStart, SEEK_SET) < 0) {
-	    rpmlog(RPMLOG_ERR, _("Could not seek in file %s: %s\n"),
-		    rpm, Fstrerror(fd));
-	    goto exit;
-	}
-
 	if (rpmWriteSignature(fd, sigh)) {
 	    rpmlog(RPMLOG_ERR, _("%s: rpmWriteSignature failed: %s\n"), rpm,
 		Fstrerror(fd));
@@ -641,6 +715,7 @@ static int rpmSign(const char *rpm, int deleting, int signfiles)
 	}
 	res = 0;
     } else {
+	char lbuf[RPMLEAD_SIZE];
 	/* Replace orignal rpm with new rpm containing new signature */
 	rasprintf(&trpm, "%s.XXXXXX", rpm);
 	ofd = rpmMkTemp(trpm);
@@ -649,11 +724,12 @@ static int rpmSign(const char *rpm, int deleting, int signfiles)
 	    goto exit;
 	}
 
-	/* Write the lead/signature of the output rpm */
-	rc = rpmLeadWrite(ofd, h);
-	if (rc != RPMRC_OK) {
-	    rpmlog(RPMLOG_ERR, _("%s: writeLead failed: %s\n"), trpm,
-		Fstrerror(ofd));
+	/* Copy lead from original */
+	if (!(Fread(lbuf, 1, RPMLEAD_SIZE, fd) == RPMLEAD_SIZE &&
+	      Fwrite(lbuf, 1, RPMLEAD_SIZE, ofd) == RPMLEAD_SIZE))
+	{
+	    rpmlog(RPMLOG_ERR, _("%s: failed to copy rpm lead: %s\n"),
+				trpm, Fstrerror(ofd));
 	    goto exit;
 	}
 
@@ -672,6 +748,10 @@ static int rpmSign(const char *rpm, int deleting, int signfiles)
 	if (copyFile(&fd, rpm, &ofd, trpm) == 0) {
 	    struct stat st;
 
+	    /* File must be closed before deletion due to different file locking in some file systems*/
+	    if (fd) (void) closeFile(&fd);
+	    if (ofd) (void) closeFile(&ofd);
+
 	    /* Move final target into place, restore file permissions. */
 	    if (stat(rpm, &st) == 0 && unlink(rpm) == 0 &&
 			rename(trpm, rpm) == 0 && chmod(rpm, st.st_mode) == 0) {
@@ -716,7 +796,7 @@ int rpmPkgSign(const char *path, const struct rpmSignArgs * args)
 	}
     }
 
-    rc = rpmSign(path, 0, args ? args->signfiles : 0);
+    rc = rpmSign(path, 0, args ? args->signflags : 0);
 
     if (args) {
 	if (args->hashalgo) {
@@ -734,3 +814,8 @@ int rpmPkgDelSign(const char *path, const struct rpmSignArgs * args)
 {
     return rpmSign(path, 1, 0);
 }
+
+int rpmPkgDelFileSign(const char *path, const struct rpmSignArgs * args)
+{
+    return rpmSign(path, 2, 0);
+}
